# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime
import json

import six

from driverng.worklist import Worklist


class GraphContainer(object):
    def __init__(self, type_fichier=None):
        self.type_fichier = type_fichier

    def get_bytes(self):
        """Méthode qui doit retourner le graphe."""
        raise NotImplementedError


class JSONGraphContainer(dict):
    @property
    def type_fichier(self):
        return "json"

    def get_bytes(self):
        if six.PY3:
            return bytes(json.dumps(self), "utf8")  # bytea needs bytes in python3
        return json.dumps(self)


class BinGraphContainer(dict):
    @property
    def type_fichier(self):
        return self["format"]

    def get_bytes(self):
        return self["data"]


class Option(object):
    option_type = "string"

    def __init__(
        self,
        name,
        label=None,
        default=None,
        choices=None,
        min_value=None,
        max_value=None,
    ):
        self.name = name
        self.label = label or name
        self.default = default
        self.choices = choices
        self.min_value = min_value
        self.max_value = max_value
        super(Option, self).__init__()

    def serialize(self):
        return {
            "type": self.option_type,
            "name": self.name,
            "label": self.label,
            "default": self.default,
            "choices": self.choices,
            "min_value": self.min_value,
            "max_value": self.max_value,
        }


class BooleanOption(Option):
    option_type = "boolean"


class IntegerOption(Option):
    option_type = "integer"


class Message(object):
    _ImplementedByChild = NotImplementedError("Must be implemented in child classes")
    QUERY, ORDER, QC, RESULT, BILLING, HEADER = list(range(6))
    ENCODING = "ISO-8859-15"

    Option = Option
    BooleanOption = BooleanOption
    IntegerOption = IntegerOption
    OPTIONS = []
    BINARY_PARSING = False

    def __init__(self, *args, **kwargs):
        super(Message, self).__init__()

    @classmethod
    def get_options(cls):
        return [o.serialize() for o in cls.OPTIONS]

    def create(self, *args, **kwargs):
        raise self._ImplementedByChild

    def parse(self, data):
        """Parse le message `data` et renvoie le message."""
        raise self._ImplementedByChild

    def get_queries(self, message):
        """Retourne la liste des tube en query."""
        return []

    def to_string(self, message, **kwargs):
        raise self._ImplementedByChild

    def get_message_type(self, message):
        raise self._ImplementedByChild

    def get_order_message(self, worklist, context):
        pass

    def get_void_order_msg(self, message, dem_auto_found=False, context=None):
        """DEPRECATED."""
        if context is None:
            context = {}
        context["query"] = message
        context.setdefault("now", datetime.datetime.now())
        return self.get_order_message(Worklist(), context)

    def get_result_dict(self, message):
        """Retourne la liste des resultats patients
        {tube: {
            rack: "rack-pos_rack",
            resultats: [{
                "test": code_test,
                "resu": resultat,
                "alarmes": liste d'alarmes,
                "commentaires": Liste de commentaires,
                "lots_reactifs": Liste de lots réactifs,
                "lots_calibrations": Liste de lots de calibration,
                "lots_controles": Liste de lots de controles,
                "date": date du resultat,
                "analyseur": Analyseur,
                "zero": [liste de valeurs qui sont à ignorer quand
                         param_auto.resu_azero vaut 0],
                "graph": Doit retourner un object GraphContainer.
                "type_prlv": Le code fluide (param_auto),
            }]
        }}
        """
        return {}

    def get_qc_result_list(self, message):
        """Retourne la liste des resultats QC
        [{
            "lot": lot,
            "test": code_test,
            "resu": resultat,
            "date": date,
            "sample_id": reference réele,
            "control_level": niveau,
            "type_prlv": type_prlv
            "lots_reactifs": Liste de lots réactifs,
            "lots_calibrations": Liste de lots de calibration,
            "lots_controles": Liste de lots de controles,
        }]
        """
        return []

    def enregistrer(self, data, message=None):
        # Cette méthode est appellée par service_auto pour savoir
        # s'il faut enregistrer les données dans la table
        # integration.
        return True

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError
        for sample in worklist.samples:
            if sample.services:
                sample.service = sample.services[0].code
            else:
                sample.service = ""
            if sample.physicians:
                sample.physician = sample.physicians[0].code
                sample.physician_name = sample.physicians[0].last_name
                sample.physician_first_name = sample.physicians[0].first_name
            else:
                sample.physician = ""
                sample.physician_name = ""
                sample.physician_first_name = ""
        return self.get_order_message(worklist, context)

    @staticmethod
    def _lots_from_legacy(r):
        lots = []
        for lot_type, legacy_lots in (
            ("reagent", r.get("lots_reactifs", [])),
            ("calibration", r.get("lots_calibrations", [])),
            ("control", r.get("lots_controles", [])),
        ):
            for lot in legacy_lots:
                if isinstance(lot, six.string_types):
                    lots.append({"type": lot_type, "lot": lot})
                else:
                    lots.append(
                        {
                            "type": lot_type,
                            "lot": lot["ref"],
                            "name": lot.get("nom"),
                            "expiration": lot.get("peremption"),
                        }
                    )
        return lots

    @staticmethod
    def _lots_to_legacy(lots):
        legacy_lots = {
            "lots_reactifs": [],
            "lots_controles": [],
            "lots_calibrations": [],
        }
        for lot in lots:
            legacy_lots[
                {
                    "reagent": "lots_reactifs",
                    "calibration": "lots_calibrations",
                    "control": "lots_controles",
                }[lot.type]
            ].append(
                {"ref": lot["lot"], "nom": lot["name"], "peremption": lot["expiration"]}
            )
        return legacy_lots

    def decode_message(self, message):
        if self.get_message_type(message) == self.QUERY:
            wk = Worklist(context="query")
            for sample_id in self.get_queries(message):
                wk.samples.add(type="patient", sample_id=sample_id)
            return wk

        result_dict = self.get_result_dict(message)
        qc_result_list = self.get_qc_result_list(message)
        wk = Worklist(context="result")
        for sample_id, values in result_dict.items():
            specimens = set([r.get("type_prlv") for r in values["resultats"]])
            if len(specimens) > 1:
                raise RuntimeError
            elif len(specimens) == 1:
                specimen = list(specimens)[0]
            else:
                specimen = None
            sample = wk.samples.add(
                sample_id=sample_id,
                rack=values.get("rack"),
                specimen=specimen,
                type="patient",
            )
            for r in values["resultats"]:
                sample.results.add(
                    code=r["test"],
                    value=r["resu"],
                    date=r.get("date"),
                    flags=r.get("alarmes", []),
                    comments=r.get("commentaires", []),
                    lots=self._lots_from_legacy(r),
                    instrument=r.get("analyseur"),
                    status=r.get("status"),
                    operator=r.get("operator", ""),
                    preanalytique=r.get("preanalytique", ""),
                )
        qcs = {}
        for r in qc_result_list:
            key = (r["lot"], r.get("sample_id"))
            if key in qcs:
                sample = qcs[key]
            else:
                sample = wk.samples.add(
                    control_lot=r["lot"],
                    control_level=r.get("control_level"),
                    control_name=r.get("nom_lot"),
                    sample_id=r.get("sample_id"),
                    specimen=r.get("type_prlv"),
                    type="control",
                )
                qcs[key] = sample
            sample.results.add(
                code=r["test"],
                value=r["resu"],
                date=r["date"],
                lots=self._lots_from_legacy(r),
                instrument=r.get("analyseur"),
                status=r.get("status"),
            )
        return wk

    def encode(self, worklist, context=None):
        """Encode une worklist en un message

        :param worklist: La Worklist à encoder
        :rtype: Le message encodé (sous forme de chaine unicode) ou None
        """
        worklist.validate()
        context = context or {}
        context.setdefault("now", datetime.datetime.now())
        context.setdefault("query", None)
        message = self.encode_worklist(worklist, context)
        if message is not None:
            if type(message) == list:
                data = ""
                for m in message:
                    data += self.to_string(m)
                return data
            else:
                return self.to_string(message)

    def decode(self, data):
        """Décode un message brut

        :param data: Le message sous forme de chaine unicode
        :rtype: driverng.Worklist
        """
        worklist = self.decode_message(self.parse(data))
        if worklist is not None:
            worklist.validate()
        return worklist


class MessageV2(Message):
    def encode_worklist(self, worklist, context):
        """Méthode à hériter, elle doit générer un objet représentant le message généré à partir de la worklist."""
        raise self._ImplementedByChild

    def decode_message(self, message):
        """Méthode à hériter pour décoder un message (objet) en une worklist"""
        raise self._ImplementedByChild

    def get_order_message(self, worklist, context):
        """DEPRECATED"""
        if len(worklist.samples) and all(
            ["control_level" in s for s in worklist.samples]
        ):
            # Message de resultat de QC envoyé par un MCA <= 6.13
            worklist.context = "result"
            for sample in worklist.samples:
                sample.type = "control"
                for result in sample.results:
                    if result.reagent:
                        result.lots.add(type="reagent", lot=result.reagent)
        else:
            worklist.context = "order"
            for sample in worklist.samples:
                if sample.service:
                    sample.services = [{"code": sample.service}]
                if sample.physician:
                    sample.physicians = [
                        {
                            "code": sample.physician,
                            "last_name": sample.physician_name,
                            "first_name": sample.physician_first_name,
                            "type": "prescriber",
                        }
                    ]
                for result in sample.results:
                    if result.get("reagent"):
                        result.lots.append(type="reagent", lot=result.reagent)
        return self.encode_worklist(worklist, context)

    def get_message_type(self, message):
        """DEPRECATED"""
        worklist = self.decode_message(message)
        if worklist is None:
            raise RuntimeError("No worklist")
        if worklist.context == "query":
            return self.QUERY
        else:
            return self.ORDER

    def get_result_dict(self, message):
        """DEPRECATED"""
        results = {}
        worklist = self.decode_message(message)
        worklist.validate()
        if worklist.context != "result":
            return {}
        for sample in worklist.samples:
            if sample.type != "patient":
                continue
            resultats = results.setdefault(
                sample.sample_id, {"rack": sample.rack, "resultats": []}
            )["resultats"]
            for r in sample.results:
                res = {
                    "test": r.code,
                    "resu": r.value,
                    "date": r.date,
                    "alarmes": r.flags,
                    "commentaires": r.comments,
                    "analyseur": r.instrument,
                    "type_prlv": sample.specimen,
                    "dilution": r.dilution,
                    "operator": r.operator,
                    "unit": r.unit,
                    "preanalytique": r.preanalytique,
                    "nom_antibiogramme": r.nom_antibiogramme,
                }
                if r.validator:
                    res["validator"] = r.validator
                # Conversion des antibiogram de la worklist -> get_result_dict()
                if (
                    isinstance(res["resu"], list)
                    and res["resu"]
                    and isinstance(res["resu"][0], dict)
                    and "antibiogram" in res["resu"][0]
                ):
                    res["atb"] = [
                        (
                            ("ANTIBIO", x.get("antibiotic")),
                            ("CMI", x.get("mic")),
                            ("SENSI", x.get("sensitivity")),
                            ("DIAMETRE", x.get("diametre")),
                            ("CUTOFF", x.get("cutoff")),
                        )
                        for x in res["resu"][0]["antibiogram"]
                    ]
                    if r.value and isinstance(r.value, list):
                        res["seq"] = r.value[0].get("sub_code")  # rang du germe
                        res["phenotypes"] = r.value[0].get(
                            "phenotypes"
                        )  # marqueurs de resiostance
                    if r.sub_code:
                        res["seq"] = r.sub_code  # rang du germe
                    res["resu"] = res["resu"][0]["value"]
                # Fin de la conversion des antibiogram

                # Conversion des resultats JSON de la worklist -> get_result_dict()
                if (
                    isinstance(res["resu"], list)
                    and res["resu"]
                    and isinstance(res["resu"][0], dict)
                    and "json" in res["resu"][0]
                ):
                    res["resu"] = res["resu"][0]["json"].copy()
                # Fin

                if r.graph:
                    if r.graph.graph_type == "binary":
                        res["graph"] = BinGraphContainer(r.graph)
                    else:
                        res["graph"] = JSONGraphContainer(r.graph)
                res.update(**self._lots_to_legacy(r["lots"]))
                resultats.append(res)
        return results

    def get_qc_result_list(self, message):
        """DEPRECATED"""
        worklist = self.decode_message(message)
        worklist.validate()
        if worklist.context != "result":
            return []

        results = []
        for sample in worklist.samples:
            if sample.type != "control":
                continue
            for r in sample.results:
                res = {
                    "lot": sample.control_lot,
                    "test": r.code,
                    "resu": r.value,
                    "date": r.date,
                }
                for key, value in (
                    ("control_name", sample.control_name),
                    ("control_level", sample.control_level),
                    ("analyseur", r.instrument),
                    ("type_prlv", sample.specimen),
                    ("flags", r.flags),
                    ("status", r.status),
                ):
                    if value:
                        res[key] = value
                res.update(**self._lots_to_legacy(r["lots"]))
                results.append(res)
        return results

    def get_queries(self, message):
        """DEPRECATED"""
        worklist = self.decode_message(message)
        queries = []
        if worklist.context == "query":
            for sample in worklist.samples:
                queries.append(sample.sample_id)
        return queries
